unit SetupEnt;

{
  Inno Setup
  Copyright (C) 1998-99 Jordan Russell
  For conditions of distribution and use, see LICENSE.TXT.

  Routines for handling records with embedded PChar strings
}

interface

uses
  zlib;

procedure SEFreeStr (var P: PChar);
procedure SEAssignStrBuf (var P: PChar; const Buf; const Count: Cardinal);
procedure SEAssignStr (var P: PChar; const S: String);
procedure SEFreeRec (const P: Pointer; const Bytes: Cardinal; const NumStrings: Integer);
procedure SEDuplicateRec (OldP, NewP: Pointer; Bytes: Cardinal;
  const NumStrings: Integer);
procedure SEDeflateBlockWrite (var Data: TDeflateBlockWriteData; var Buf;
  const Count: Cardinal; const NumStrings: Integer);
procedure SEInflateBlockRead (var Data: TDeflateBlockReadData; var Buf;
  const Count: Cardinal; const NumStrings: Integer);

implementation

procedure SEFreeStr (var P: PChar);
var
  P2: Pointer;
  Len: Longint;
begin
  if P = nil then Exit;
  P2 := P;
  Dec (Cardinal(P2), SizeOf(Longint));
  Len := Longint(P2^);
  FreeMem (P2, Len + SizeOf(Longint)+1);
  P := nil;
end;

procedure SEAssignStrBuf (var P: PChar; const Buf; const Count: Cardinal);
var
  P2: Pointer;
begin
  SEFreeStr (P);
  GetMem (P2, Count + SizeOf(Longint)+1);
  Longint(P2^) := Count;
  Inc (Cardinal(P2), SizeOf(Longint));
  P := P2;
  Move (Buf, P2^, Count);
  Inc (Cardinal(P2), Count);
  Byte(P2^) := 0;
end;

procedure SEAssignStr (var P: PChar; const S: String);
begin
  SEAssignStrBuf (P, {$IFDEF WIN32} PChar(S)^ {$ELSE} S[1] {$ENDIF}, Length(S));
end;

procedure SEFreeRec (const P: Pointer; const Bytes: Cardinal; const NumStrings: Integer);
var
  P2: Pointer;
  I: Integer;
begin
  if P = nil then Exit;
  P2 := P;
  for I := 1 to NumStrings do begin
    SEFreeStr (PChar(P2^));
    Inc (Cardinal(P2), SizeOf(Pointer));
  end;
  FreeMem (P, Bytes);
end;

procedure SEDuplicateRec (OldP, NewP: Pointer; Bytes: Cardinal;
  const NumStrings: Integer);
var
  I: Integer;
  P: Pointer;
begin
  FillChar (NewP^, NumStrings * SizeOf(Pointer), 0);
  for I := 1 to NumStrings do begin
    if Assigned(PChar(OldP^)) then begin
      P := Pointer(OldP^);
      Dec (Cardinal(P), SizeOf(Longint));
      SEAssignStrBuf (PChar(NewP^), Pointer(OldP^)^, Longint(P^));
    end;
    Inc (Cardinal(OldP), SizeOf(Pointer));
    Inc (Cardinal(NewP), SizeOf(Pointer));
    Dec (Bytes, SizeOf(Pointer));
  end;
  Move (OldP^, NewP^, Bytes);
end;

procedure SEDeflateBlockWrite (var Data: TDeflateBlockWriteData; var Buf;
  const Count: Cardinal; const NumStrings: Integer);
var
  TotalSize: Cardinal;
  P, PL, PL2, P2: Pointer;
  I: Integer;
  NewBuf: Pointer;
  Size: Cardinal;
  TotalSizeLong: Longint;
begin
  TotalSize := Count;
  P := @Buf;
  for I := 1 to NumStrings do begin
    PL := Pointer(P^);
    if PL <> nil then begin
      Dec (Cardinal(PL), SizeOf(Longint));
      Inc (TotalSize, Longint(PL^));
    end;
    Inc (Cardinal(P), SizeOf(Pointer));
  end;
  GetMem (NewBuf, TotalSize);
  try
    P := @Buf;
    P2 := NewBuf;
    for I := 1 to NumStrings do begin
      PL := Pointer(P^);
      if PL <> nil then begin
        PL2 := PL;
        Dec (Cardinal(PL2), SizeOf(Longint));
        Size := Longint(PL2^);
        Longint(P2^) := Size;
        Inc (Cardinal(P2), SizeOf(Longint));
        Move (PL^, P2^, Size);
        Inc (Cardinal(P2), Size);
      end
      else begin
        Longint(P2^) := 0;
        Inc (Cardinal(P2), SizeOf(Longint));
      end;
      Inc (Cardinal(P), SizeOf(Pointer));
    end;
    P := @Buf;
    Inc (Cardinal(P), NumStrings * SizeOf(Pointer));
    Move (P^, P2^, Count - (Cardinal(NumStrings) * SizeOf(Pointer)));
    TotalSizeLong := TotalSize;
    DeflateBlockWrite (Data, TotalSizeLong, SizeOf(TotalSizeLong));
    DeflateBlockWrite (Data, NewBuf^, TotalSize);
  finally
    FreeMem (NewBuf, TotalSize);
  end;
end;

procedure SEInflateBlockRead (var Data: TDeflateBlockReadData; var Buf;
  const Count: Cardinal; const NumStrings: Integer);
var
  UncompressedSize: Longint;
  TmpBuf, InP, OutP: Pointer;
  I: Integer;
  Len: Longint;
begin
  InflateBlockRead (Data, UncompressedSize, SizeOf(UncompressedSize));
  GetMem (TmpBuf, UncompressedSize);
  try
    InflateBlockRead (Data, TmpBuf^, UncompressedSize);
    InP := TmpBuf;
    OutP := @Buf;
    for I := 1 to NumStrings do begin
      Len := Longint(InP^);
      Inc (Cardinal(InP), SizeOf(Longint));
      PChar(OutP^) := nil;
      SEAssignStrBuf (PChar(OutP^), InP^, Len);
      Inc (Cardinal(InP), Len);
      Inc (Cardinal(OutP), SizeOf(Pointer));
    end;
    Move (InP^, OutP^, Cardinal(UncompressedSize) - (Cardinal(InP) - Cardinal(TmpBuf)));
  finally
    FreeMem (TmpBuf, UncompressedSize);
  end;
end;

end.

