unit Undo;

{
  Inno Setup
  Copyright (C) 1998-99 Jordan Russell
  For conditions of distribution and use, see LICENSE.TXT.

  Uninstallation Procedures
}

{ Note: This unit is shared by both the 'Setup' and 'Uninst' projects }

interface

{$I VERSION.INC}

uses
  WinProcs, WinTypes, SysUtils;

type
  { To figure out how to use these, look for "UndoChanges.Add" throughout the
    Install unit of SETUP }
  TUndoChangeTyp = (utRun, utDeleteDirOrFiles, utDeleteFile,
    utDeleteGroupOrItem, utIniDeleteEntry, utIniDeleteSection,
    utRegDeleteEntireKey, utRegClearValue
    {$IFDEF WIN32}, utRegDeleteKeyIfEmpty, utRegDeleteValue,
      utDecrementSharedCount {$ENDIF});
const
  utRun_NoWait = 1;
  utRun_WaitUntilIdle = 2;
  utRun_ShellExec = 4;
  utDeleteFile_ExistedBeforeInstall = 1;
  utDeleteFile_Extra = 2;
  utDeleteFile_IsFont = 4;
  {$IFDEF WIN32}
  utDeleteFile_SharedFile = 8;
  utDeleteFile_RegisteredServer = 16;
  utDeleteFile_CallChangeNotify = 32;
  utDeleteFile_RegisteredTypeLib = 64;
  {$ENDIF}
  utDeleteDirOrFiles_Extra = 1;
  utDeleteDirOrFiles_IsDir = 2;
  utDeleteDirOrFiles_DeleteFiles = 4;
  utDeleteDirOrFiles_DeleteSubdirsAlso = 8;
  {$IFDEF WIN32}
  utDeleteDirOrFiles_CallChangeNotify = 16;
  {$ENDIF}
  utIniDeleteSection_OnlyIfEmpty = 1;

  UndoChangeRecNonDataSize =
    (SizeOf(Pointer) * 2) + SizeOf(Longint) + SizeOf(TUndoChangeTyp) + SizeOf(Word);
  MaxUndoDataSize = 65000 - UndoChangeRecNonDataSize;
type
  PUndoChangeRec = ^TUndoChangeRec;
  TUndoChangeRec = packed record
    {if this is modified, you must to update UndoChangeRecNonDataSize above}
    Prev, Next: PUndoChangeRec;
    ExtraData: Longint;
    Typ: TUndoChangeTyp;
    DataSize: Word;
    Data: array[0..MaxUndoDataSize-1] of Char;
  end;

type
  TDeleteUninstallDataFilesProc = procedure;

  TUndoChanges = class
  private
    List, LastList: PUndoChangeRec;
  public
    AppName, AppDir: String;
    AdminInstalled: Boolean;
    destructor Destroy; override;
    procedure AddP (const Typ: TUndoChangeTyp; const Data: Pointer; const Size: Word;
      const ExtraData: Longint);
    procedure Add (const Typ: TUndoChangeTyp; const Data: array of String;
      const ExtraData: Longint);
    procedure Clear;
    function  PerformUndo (const CallFromUninstaller: Boolean;
      const DeleteUninstallDataFilesProc: TDeleteUninstallDataFilesProc): Boolean;
    procedure Save (const Filename: String);
    procedure Load (const Filename: String);
  end;

implementation

uses
  Messages,
  CmnFunc2,
  Struct, Msgs, MsgIDs,
  ShellAPI,
  {$IFDEF WIN32} ShlObj, {$IFNDEF Delphi3orHigher} OLE2, {$ELSE} ActiveX, ComObj, {$ENDIF} {$ENDIF}
  DDEInt, InstFunc;

type
  TUndoFileHeader = record
    ID: TUndoID;
    NumUndos: Cardinal;
    AppName: String[63];
    AppDir: array[0..259] of Char;
    AdminInstalled: Boolean;
  end;

destructor TUndoChanges.Destroy;
begin
  Clear;
  inherited Destroy;
end;

procedure TUndoChanges.AddP (const Typ: TUndoChangeTyp; const Data: Pointer;
  const Size: Word; const ExtraData: Longint);
{ Adds a new entry to the uninstall list }
var
  NewRec: PUndoChangeRec;
begin
  NewRec := AllocMem(UndoChangeRecNonDataSize + Size);
  NewRec^.ExtraData := ExtraData;
  NewRec^.Typ := Typ;
  NewRec^.DataSize := Size;
  if NewRec^.DataSize > MaxUndoDataSize then
    NewRec^.DataSize := MaxUndoDataSize;
  Move (Data^, NewRec^.Data, Size);
  if List = nil then begin
    List := NewRec;
    LastList := List;
  end
  else begin
    LastList^.Next := NewRec;
    NewRec^.Prev := LastList;
    LastList := NewRec;
  end;
end;

procedure TUndoChanges.Add (const Typ: TUndoChangeTyp; const Data: array of String;
  const ExtraData: Longint);
{ Same as AddP, but accepts string type }
var
  I: Integer;
  {$IFDEF WIN32}
  S: String;
  {$ELSE}
  Buf, P: PChar;
  Size: Word;
  {$ENDIF}
begin
  {$IFDEF WIN32}
  for I := 0 to High(Data) do
    S := S + Data[I] + #0;
  AddP (Typ, PChar(S), Length(S), ExtraData);
  {$ELSE}
  GetMem (Buf, 256 * (High(Data)+1));
  P := Buf;
  Size := 0;
  for I := 0 to High(Data) do begin
    Move (Data[I, 1], P^, Length(Data[I]));
    Inc (P, Length(Data[I]));
    P^ := #0;
    Inc (P);
    Inc (Size, Length(Data[I])+1);
  end;
  AddP (Typ, Buf, Size, ExtraData);
  FreeMem (Buf, 256 * (High(Data)+1));
  {$ENDIF}
end;

procedure TUndoChanges.Clear;
{ Frees all entries in the uninstall list and clears AppName/AppDir }
var
  CurRec, NextRec: PUndoChangeRec;
begin
  CurRec := List;
  while CurRec <> nil do begin
    NextRec := CurRec^.Next;
    FreeMem (CurRec {$IFNDEF WIN32} , UndoChangeRecNonDataSize + CurRec^.DataSize {$ENDIF});
    CurRec := NextRec;
  end;
  List := nil;
  LastList := nil;
  AppName := '';
  AppDir := '';
  AdminInstalled := False;
end;

{$IFDEF WIN32}
function DecrementSharedCount (const Filename: String): Boolean;
{ Returns True if OK to delete }
const
  SharedDLLsKey = NEWREGSTR_PATH_SETUP + '\SharedDLLs';  {don't localize}
var
  K: HKEY;
  Count, CurType, NewType, Size, Disp: DWORD;
  CountStr: String;
begin
  Result := False;

  if RegCreateKeyEx(HKEY_LOCAL_MACHINE, SharedDLLsKey, 0, nil, REG_OPTION_NON_VOLATILE,
     KEY_QUERY_VALUE or KEY_SET_VALUE, nil, K, @Disp) <> ERROR_SUCCESS then
    raise Exception.Create(FmtSetupMessage(msgErrorRegOpenKey,
      [RegRootKeyNames[HKEY_LOCAL_MACHINE], SharedDLLsKey]));
  if RegQueryValueEx(K, PChar(Filename), nil, @CurType, nil, @Size) <> ERROR_SUCCESS then begin
    RegCloseKey (K);
    Exit;
  end;

  Count := 2;
  NewType := REG_DWORD;
  try
    case CurType of
      REG_SZ:
        if RegQueryStringValue(K, PChar(Filename), CountStr) then begin
          Count := StrToInt(CountStr);
          NewType := REG_SZ;
        end;
      REG_BINARY: begin
          if (Size >= 1) and (Size <= 4) then begin
            if RegQueryValueEx(K, PChar(Filename), nil, nil, @Count, @Size) <> ERROR_SUCCESS then
              { ^ relies on the high 3 bytes of Count being initialized to 0 }
              Abort;
            NewType := REG_BINARY;
          end;
        end;
      REG_DWORD: begin
          Size := SizeOf(DWORD);
          if RegQueryValueEx(K, PChar(Filename), nil, nil, @Count, @Size) <> ERROR_SUCCESS then
            Abort;
        end;
    end;
  except
    Count := 2;  { default to 2 if an error occurred }
  end;
  Dec (Count);
  if Count <= 0 then begin
    Result := True;
    RegDeleteValue (K, PChar(Filename));
  end
  else begin
    case NewType of
      REG_SZ: begin
          CountStr := IntToStr(Count);
          RegSetValueEx (K, PChar(Filename), 0, NewType, PChar(CountStr), Length(CountStr)+1);
        end;
      REG_BINARY, REG_DWORD:
        RegSetValueEx (K, PChar(Filename), 0, NewType, @Count, SizeOf(Count));
    end;
  end;
  RegCloseKey (K);
end;

function UnregisterServer (const Filename: String): Boolean;
var
  LibHandle: THandle;
  UnregisterServerProc: function: HRESULT; stdcall;
begin
  Result := True;
  try
    LibHandle := LoadLibrary(PChar(Filename));
    if LibHandle <> 0 then begin
      try
        @UnregisterServerProc := GetProcAddress(LibHandle, 'DllUnregisterServer');
        if Assigned(@UnregisterServerProc) and SUCCEEDED(UnregisterServerProc) then
          Exit;
      finally
        FreeLibrary (LibHandle);
      end;
    end;
  except
  end;
  Result := False;
end;

function UnregisterTypeLib (const Filename: String): Boolean;
type
  TUnRegTlbProc = function (const libID: TGUID; wVerMajor, wVerMinor: Word;
    lcid: TLCID; syskind: TSysKind): HResult; stdcall;
{$IFNDEF Delphi3OrHigher}
var
  WideFilename: PWideChar;
  TypeLib: ITypeLib;
  OleAutLib: THandle;
  UnRegTlbProc: TUnRegTlbProc;
  LibAttr: PTLibAttr;
begin
  Result := False;
  WideFilename := nil;
  TypeLib := nil;
  LibAttr := nil;
  try
    { Dynamically import UnRegisterTypeLib since older OLEAUT32.DLL versions
      don't have this function }
    OleAutLib := GetModuleHandle('OLEAUT32.DLL');
    if OleAutLib = 0 then
      Exit;
    @UnRegTlbProc := GetProcAddress(OleAutLib, 'UnRegisterTypeLib');;
    if @UnRegTlbProc = nil then
      Exit;
    WideFilename := StringToOleStr(ExpandFilename(Filename));
    if FAILED(LoadTypeLib(WideFilename, TypeLib)) then
      Exit;
    if FAILED(TypeLib.GetLibAttr(LibAttr)) then
      Exit;
    with LibAttr^ do
      if FAILED(UnRegTlbProc(Guid, wMajorVerNum, wMinorVerNum, LCID, SysKind)) then
        Exit;
  finally
    if Assigned(TypeLib) then begin
      if Assigned(LibAttr) then
        TypeLib.ReleaseTLibAttr (LibAttr);
      TypeLib.Release;
    end;
    if Assigned(WideFilename) then
      SysFreeString (WideFilename);
  end;
  Result := True;
end;
{$ELSE}
var
  WideFilename: WideString;
  TypeLib: ITypeLib;
  OleAutLib: THandle;
  UnRegTlbProc: TUnRegTlbProc;
  LibAttr: PTLibAttr;
begin
  Result := False;
  { Dynamically import UnRegisterTypeLib since older OLEAUT32.DLL versions
    don't have this function }
  OleAutLib := GetModuleHandle('OLEAUT32.DLL');
  if OleAutLib = 0 then
    Exit;
  @UnRegTlbProc := GetProcAddress(OleAutLib, 'UnRegisterTypeLib');;
  if @UnRegTlbProc = nil then
    Exit;
  WideFilename := ExpandFilename(Filename);
  if FAILED(LoadTypeLib(PWideChar(WideFilename), TypeLib)) then
    Exit;
  if FAILED(TypeLib.GetLibAttr(LibAttr)) then
    Exit;
  try
    with LibAttr^ do
      if FAILED(UnRegTlbProc(Guid, wMajorVerNum, wMinorVerNum, LCID, SysKind)) then
        Exit;
  finally
    TypeLib.ReleaseTLibAttr (LibAttr);
  end;
  Result := True;
end;
{$ENDIF}
{$ENDIF}

function DeleteDirWait (const DirName: String;
  const DirsRemoved, DirsNotRemoved: TSimpleStringList): Boolean;
{$IFDEF WIN32}
var
  I, J: Integer;
  FindHandle: THandle;
  FindData: TWin32FindData;
  DidntGetRemoved: Boolean;
{$ENDIF}
begin
  {$IFDEF WIN32}
  { Sometimes the IE4 shell has shell folders locked, so wait
    up to 5 seconds for previously removed subdirectories to
    *actually* be deleted. }
  for I := 0 to 19 do begin
    DidntGetRemoved := False;
    for J := 0 to DirsRemoved.Count-1 do
      if Copy(DirsRemoved[J], 1, Length(DirName)+1) =
         DirName + '\' then begin
        FindHandle := FindFirstFile(PChar(DirsRemoved[J]), FindData);
        if FindHandle <> INVALID_HANDLE_VALUE then begin
          WinProcs.FindClose (FindHandle);
          if FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY <> 0 then begin
            DidntGetRemoved := True;
            Break;
          end;
        end;
      end;
    if not DidntGetRemoved then Break;
    Sleep (250);
  end;
  Result := RemoveDirectory(PChar(DirName));
  if Result then
    DirsRemoved.AddIfDoesntExist (DirName)
  else
    if Assigned(DirsNotRemoved) then
      DirsNotRemoved.AddIfDoesntExist (DirName);
  {$ELSE}
  Result := True;
  try
    RmDir (DirName);
  except
    Result := False;
    if Assigned(DirsNotRemoved) then
      DirsNotRemoved.AddIfDoesntExist (DirName);
  end;
  {$ENDIF}
end;

type
  PDeleteDirData = ^TDeleteDirData;
  TDeleteDirData = record
    DirsRemoved, DirsNotRemoved: TSimpleStringList;
  end;

function DeleteDirProc (const DirName: String; Param: Pointer): Boolean; far;
begin
  Result := DeleteDirWait(DirName, PDeleteDirData(Param)^.DirsRemoved,
    PDeleteDirData(Param)^.DirsNotRemoved);
end;

procedure ProcessMessagesProc; far;
var
  Msg: TMsg;
begin
  while PeekMessage(Msg, 0, 0, 0, PM_REMOVE) do begin
    TranslateMessage (Msg);
    DispatchMessage (Msg);
  end;
end;

function TUndoChanges.PerformUndo (const CallFromUninstaller: Boolean;
  const DeleteUninstallDataFilesProc: TDeleteUninstallDataFilesProc): Boolean;
{ Undoes all the changes in the uninstall list, in reverse order they were
  added. Deletes entries that were successfully undone.
  Returns True if all elements were successfully removed; False if some
  could not be removed. }

var
  {$IFDEF WIN32}
  ChangeNotifyList: TSimpleStringList;
  {$ENDIF}
  DeleteDirData: TDeleteDirData;

  procedure FileDelete (const Filename: String {$IFDEF WIN32} ; const NotifyChange: Boolean {$ENDIF});
  begin
    if CompareText(ExtractFileExt(Filename), '.HLP') = 0 then begin
      DeleteFile (ChangeFileExt(Filename, '.GID'));
      DeleteFile (ChangeFileExt(Filename, '.FTS'));
    end;
    if FileExists(Filename) then begin
      if not DeleteFile(Filename) then
        Result := False
      {$IFDEF WIN32}
      else
        if NotifyChange then begin
          SHChangeNotify (SHCNE_DELETE, SHCNF_PATH, PChar(Filename), nil);
          ChangeNotifyList.AddIfDoesntExist (RemoveBackslashUnlessRoot(ExtractFilePath(Filename)));
        end;
      {$ENDIF}
    end;
  end;

  function IsSectionEmpty (const Section, Filename: PChar): Boolean;
  var
    Test: array[0..255] of Char;
  begin
    Test[0] := #0;
    if Filename^ <> #0 then
      GetPrivateProfileString (Section, nil, '', Test, SizeOf(Test), Filename)
    else
      GetProfileString (Section, nil, '', Test, SizeOf(Test));
    Result := Test[0] = #0;
  end;
const
  GroupInfoChars: array[0..3] of Char = ('"', '"', ',', ',');
  NullChar: Char = #0;
var
  CurRec, NextRec: PUndoChangeRec;
  CurRecData, FN: String;
  P: Integer;
  K: HKEY;
  {$IFDEF WIN32}
  NumSubkeys, NumValues: DWORD;
  {$ENDIF}

  PROGMAN, Item: TDDEString;
  Conv: TDDEConversation;
  ItemName: String;
  DataHandle: TDDEData;
  DataPointer: PChar;
  DataSize: Longint;
  NoItemsLeft: Boolean;
  Msg: String;

  function GetDataIndexPChar (const Index: Integer): PChar;
  var
    I: Integer;
  begin
    Result := CurRec^.Data;
    for I := 1 to Index do
      Inc (Result, StrLen(Result)+1);
  end;
  function GetDataIndex (const Index: Integer): String;
  begin
    Result := {$IFNDEF WIN32}StrPas({$ENDIF} GetDataIndexPChar(Index) {$IFNDEF WIN32}){$ENDIF};
  end;
begin
  Result := True;

  PROGMAN := 0;
  Conv := 0;
  {$IFDEF WIN32}
  ChangeNotifyList := TSimpleStringList.Create;
  DeleteDirData.DirsRemoved := TSimpleStringList.Create;
  {$ELSE}
  DeleteDirData.DirsRemoved := nil;
  {$ENDIF}
  if Assigned(DeleteUninstallDataFilesProc) then
    DeleteDirData.DirsNotRemoved := TSimpleStringList.Create
  else
    DeleteDirData.DirsNotRemoved := nil;
  try
    CurRec := LastList;
    while CurRec <> nil do begin
      NextRec := CurRec^.Prev;

      CurRecData := StrPas(CurRec^.Data);

      try
        case CurRec^.Typ of
          utRun: begin
              if CallFromUninstaller then
                if CurRec^.ExtraData and utRun_ShellExec = 0 then begin
                  if not InstExec(CurRecData, GetDataIndex(1), GetDataIndex(2),
                     CurRec^.ExtraData and utRun_NoWait = 0,
                     CurRec^.ExtraData and utRun_WaitUntilIdle <> 0,
                     ProcessMessagesProc, P) then
                    Result := False;
                end
                else begin
                  if not InstShellExec(CurRecData, GetDataIndex(1), GetDataIndex(2), P) then
                    Result := False;
                end;
            end;
          utDeleteDirOrFiles:
            if (CallFromUninstaller or (CurRec^.ExtraData and utDeleteDirOrFiles_Extra = 0)) then begin
              { Since NT cannot remove the current directory of a drive,
                and there is a chance that the application directory
                could be the current directory, change to the Windows
                directory to prevent failure. }
              {$IFNDEF WIN32}
              SetCurrentDir (GetWinDir);
              {$ELSE}
              SetCurrentDirectory (PChar(GetWinDir));
              {$ENDIF}
              if DelTree(CurRecData, CurRec^.ExtraData and utDeleteDirOrFiles_IsDir <> 0,
                 CurRec^.ExtraData and utDeleteDirOrFiles_DeleteFiles <> 0,
                 CurRec^.ExtraData and utDeleteDirOrFiles_DeleteSubdirsAlso <> 0,
                 DeleteDirProc, @DeleteDirData) then begin
                {$IFDEF WIN32}
                if (CurRec^.ExtraData and utDeleteDirOrFiles_IsDir <> 0) and
                   (CurRec^.ExtraData and utDeleteDirOrFiles_CallChangeNotify <> 0) then begin
                  SHChangeNotify (SHCNE_RMDIR, SHCNF_PATH, PChar(CurRecData), nil);
                  ChangeNotifyList.AddIfDoesntExist (RemoveBackslashUnlessRoot(ExtractFilePath(CurRecData)));
                end;
                {$ENDIF}
              end;
            end;
          utDeleteFile: begin
              FN := GetDataIndex(1);
              if CallFromUninstaller or (FN = '') then
                FN := CurRecData;
              if CallFromUninstaller or (CurRec^.ExtraData and utDeleteFile_ExistedBeforeInstall = 0) then begin
                {$IFDEF WIN32}
                if (CurRec^.ExtraData and utDeleteFile_SharedFile = 0) or
                   (DecrementSharedCount(CurRecData) and FileExists(FN) and
                    (not CallFromUninstaller or
                     (MessageBox(0, PChar(FmtSetupMessage1(msgConfirmDeleteSharedFile, CurRecData)),
                      SetupMessages[msgConfirmDeleteSharedFileTitle], MB_YESNO or MB_ICONQUESTION) = IDYES))) then begin
                  if (CurRec^.ExtraData and utDeleteFile_RegisteredServer <> 0) and
                     (FN = CurRecData) then
                    UnregisterServer (CurRecData);
                  if (CurRec^.ExtraData and utDeleteFile_RegisteredTypeLib <> 0) and
                     (FN = CurRecData) then
                    UnregisterTypeLib (CurRecData);
                {$ENDIF}
                  if CurRec^.ExtraData and utDeleteFile_IsFont <> 0 then begin
                    {$IFDEF WIN32}
                    if Win32Platform <> VER_PLATFORM_WIN32_WINDOWS then begin
                    {$ENDIF}
                      WriteProfileString ('Fonts', GetDataIndexPChar(2), nil);
                    {$IFDEF WIN32}
                    end
                    else begin
                      if RegOpenKeyEx(HKEY_LOCAL_MACHINE, NEWREGSTR_PATH_SETUP + '\Fonts',
                         0, KEY_SET_VALUE, K) <> ERROR_SUCCESS then
                        Result := False
                      else begin
                        RegDeleteValue (K, GetDataIndexPChar(2));
                        RegCloseKey (K);
                      end;
                    end;
                    {$ENDIF}
                    if RemoveFontResource(GetDataIndexPChar(3)) then
                      {$IFDEF WIN32} SendNotifyMessage {$ELSE} SendMessage {$ENDIF}
                        (HWND_BROADCAST, WM_FONTCHANGE, 0, 0);
                    {$IFNDEF WIN32}
                    if CompareText(ExtractFileExt(GetDataIndex(3)), '.FOT') = 0 then
                      DeleteFile (ChangeFileExt(FN, '.FOT'));
                    {$ENDIF}
                  end;
                  if CallFromUninstaller or (CurRec^.ExtraData and utDeleteFile_Extra = 0) then
                    FileDelete (FN {$IFDEF WIN32}, CurRec^.ExtraData and utDeleteFile_CallChangeNotify <> 0 {$ENDIF});
                {$IFDEF WIN32}
                end;
                {$ENDIF}
              end
              else begin
                {$IFDEF WIN32}
                if CurRec^.ExtraData and utDeleteFile_SharedFile <> 0 then
                  DecrementSharedCount (CurRecData);
                {$ENDIF}
                if FN <> CurRecData then
                  FileDelete (FN {$IFDEF WIN32}, CurRec^.ExtraData and utDeleteFile_CallChangeNotify <> 0 {$ENDIF});
              end;
            end;
          utDeleteGroupOrItem: begin
              try
                if PROGMAN = 0 then PROGMAN := DDE.CreateString('PROGMAN'); {don't localize}
                if Conv = 0 then Conv := DDE.BeginConnection(PROGMAN, PROGMAN);
                ItemName := GetDataIndex(1);
                if ItemName <> '' then
                  { Delete just the item }
                  {don't localize}
                  DDE.Execute (Conv, '[ShowGroup("' + CurRecData + '",5)]' +
                    '[DeleteItem("' + ItemName + '")]', True)
                else begin
                  { Delete the group if it's empty }
                  Item := DDE.CreateString(CurRecData);
                  try
                    DDE.RequestBegin (Conv, Item, CF_TEXT, DataHandle, Pointer(DataPointer), DataSize);
                    try
                      for P := 0 to 3 do begin
                        DataPointer := StrScan(DataPointer, GroupInfoChars[P]);
                        if DataPointer = nil then Abort; {exception is trapped below}
                        Inc (DataPointer);
                      end;
                      NoItemsLeft := DataPointer[0] = '0';
                    finally
                      DDE.RequestEnd (DataHandle);
                    end;
                  finally
                    DDE.FreeString (Item);
                  end;
                  if NoItemsLeft then
                    {don't localize}
                    DDE.Execute (Conv, '[DeleteGroup("' + CurRecData + '")]', True);
                end;
              except
              end;
            end;
          utIniDeleteEntry: begin
              if CurRec^.Data[0] <> #0 then
                WritePrivateProfileString (GetDataIndexPChar(1), GetDataIndexPChar(2),
                  nil, CurRec^.Data)
              else
                WriteProfileString (GetDataIndexPChar(1), GetDataIndexPChar(2), nil);
            end;
          utIniDeleteSection: begin
              if (CurRec^.ExtraData and utIniDeleteSection_OnlyIfEmpty = 0) or
                 IsSectionEmpty(GetDataIndexPChar(1), CurRec^.Data) then begin
                if CurRec^.Data[0] <> #0 then
                  WritePrivateProfileString (GetDataIndexPChar(1), nil, nil,
                    CurRec^.Data)
                else
                  WriteProfileString (GetDataIndexPChar(1), nil, nil);
              end;
            end;
          utRegDeleteEntireKey:
            {$IFDEF WIN32}
            if RegOpenKeyEx(CurRec^.ExtraData, CurRec^.Data, 0, KEY_QUERY_VALUE, K) = ERROR_SUCCESS then begin
              RegCloseKey (K);
              if not RegDeleteKeyIncludingSubkeys(CurRec^.ExtraData, CurRec^.Data) then
                Result := False;
            end;
            {$ELSE}
            if RegOpenKey(HKEY_CLASSES_ROOT, CurRec^.Data, K) = ERROR_SUCCESS then begin
              RegCloseKey (K);
              if RegDeleteKey(HKEY_CLASSES_ROOT, CurRec^.Data) <> ERROR_SUCCESS then
                Result := False;
            end;
            {$ENDIF}
          utRegClearValue:
            {$IFDEF WIN32}
            if RegOpenKeyEx(CurRec^.ExtraData, CurRec^.Data, 0, KEY_SET_VALUE, K) = ERROR_SUCCESS then begin
              if RegSetValueEx(K, GetDataIndexPChar(1), 0, REG_SZ, @NullChar,
                 SizeOf(NullChar)) <> ERROR_SUCCESS then
                Result := False;
              RegCloseKey (K);
            end;
            {$ELSE}
            if RegOpenKey(HKEY_CLASSES_ROOT, CurRec^.Data, K) = ERROR_SUCCESS then begin
              if RegSetValue(K, nil, REG_SZ, @NullChar, SizeOf(NullChar)) <> ERROR_SUCCESS then
                Result := False;
              RegCloseKey (K);
            end;
            {$ENDIF}
          {$IFDEF WIN32}
          utRegDeleteKeyIfEmpty: begin
              if RegOpenKeyEx(CurRec^.ExtraData, CurRec^.Data, 0,
                 KEY_QUERY_VALUE, K) = ERROR_SUCCESS then begin
                if RegQueryInfoKey(K, nil, nil, nil, @NumSubkeys, nil, nil,
                   @NumValues, nil, nil, nil, nil) <> ERROR_SUCCESS then begin
                  RegCloseKey (K);
                  Result := False;
                end
                else begin
                  RegCloseKey (K);
                  if (NumSubkeys = 0) and (NumValues = 0) then
                    if not RegDeleteKeyIncludingSubkeys(CurRec^.ExtraData, CurRec^.Data) then
                      Result := False;
                end;
              end;
            end;
          utRegDeleteValue: begin
              if RegOpenKeyEx(CurRec^.ExtraData, CurRec^.Data, 0,
                 KEY_QUERY_VALUE or KEY_ENUMERATE_SUB_KEYS or KEY_SET_VALUE,
                 K) = ERROR_SUCCESS then begin
                if RegValueExists(K, GetDataIndexPChar(1)) and
                   (RegDeleteValue(K, GetDataIndexPChar(1)) <> ERROR_SUCCESS) then
                  Result := False;
              end;
            end;
          utDecrementSharedCount:
            DecrementSharedCount (CurRecData);
          {$ENDIF}
        end;
      except
        on E: Exception do begin
          Result := False;
          if not(E is EAbort) then begin
            Msg := E.Message + '.';
            MessageBox (0, StringAsPChar(Msg), SetupMessages[msgUninstallAppTitle], MB_OK or MB_ICONSTOP);
          end;
        end;
      end;

      FreeMem (CurRec {$IFNDEF WIN32} , UndoChangeRecNonDataSize + CurRec^.DataSize {$ENDIF});
      CurRec := NextRec;
      LastList := CurRec;
    end;
    List := nil;

    if Assigned(DeleteUninstallDataFilesProc) then begin
      DeleteUninstallDataFilesProc;
      { Now that uninstall data is deleted, try removing the directories it
        was in that couldn't be deleted before. }
      for P := 0 to DeleteDirData.DirsNotRemoved.Count-1 do
        DeleteDirWait (DeleteDirData.DirsNotRemoved[P],
          DeleteDirData.DirsRemoved, nil);
    end;
  finally
    DeleteDirData.DirsNotRemoved.Free;
    {$IFDEF WIN32}
    DeleteDirData.DirsRemoved.Free;
    for P := 0 to ChangeNotifyList.Count-1 do
      if DirExists(ChangeNotifyList[P]) then
        SHChangeNotify (SHCNE_UPDATEDIR, SHCNF_PATH or SHCNF_FLUSH,
          PChar(ChangeNotifyList[P]), nil);
    ChangeNotifyList.Free;
    {$ENDIF}
    if Conv <> 0 then DDE.EndConnection (Conv);
    if PROGMAN <> 0 then DDE.FreeString (PROGMAN);
  end;
end;

procedure TUndoChanges.Save (const Filename: String);
{ Saves all undo data to Filename }
var
  F: File;
  Header: TUndoFileHeader;
  CurRec: PUndoChangeRec;
begin
  AssignFile (F, Filename);
  FileMode := fmOpenWrite or fmShareExclusive;
  Rewrite (F, 1);
  try
    FillChar (Header, SizeOf(Header), 0);
    BlockWrite (F, Header, SizeOf(Header));
    {goes back and fills in correct values later}

    CurRec := List;
    while CurRec <> nil do begin
      BlockWrite (F, CurRec^, UndoChangeRecNonDataSize + CurRec^.DataSize);
      Inc (Header.NumUndos);

      CurRec := CurRec^.Next;
    end;

    Seek (F, 0);
    Header.ID := UndoID;
    Header.AppName := AppName;
    StrPLCopy (Header.AppDir, AppDir, SizeOf(Header.AppDir)-1);
    Header.AdminInstalled := AdminInstalled;
    BlockWrite (F, Header, SizeOf(Header));
  finally
    CloseFile (F);
  end;
end;

procedure TUndoChanges.Load (const Filename: String);
{ Loads all undo data from Filename }
var
  F: File;
  Header: TUndoFileHeader;
  Rec: PUndoChangeRec;
  I: Integer;
begin
  AssignFile (F, Filename);
  FileMode := fmOpenRead or fmShareDenyWrite;
  Reset (F, 1);
  try
    BlockRead (F, Header, SizeOf(Header));
    if Header.ID <> UndoID then
      raise Exception.Create(FmtSetupMessage1(msgUninstallDataCorrupted, Filename));
    AppName := Header.AppName;
    AppDir := StrPas(Header.AppDir);
    AdminInstalled := Header.AdminInstalled;

    New (Rec);
    try
      for I := 1 to Header.NumUndos do begin
        BlockRead (F, Rec^, UndoChangeRecNonDataSize);
        BlockRead (F, Rec^.Data, Rec^.DataSize);
        AddP (Rec^.Typ, @Rec^.Data, Rec^.DataSize, Rec^.ExtraData);
      end;
    finally
      Dispose (REc);
    end;
  finally
    CloseFile (F);
  end;
end;

{$IFDEF WIN32}
procedure InitOle;
var
  OleResult: HRESULT;
begin
  OleResult := CoInitialize(nil);
  if FAILED(OleResult) then
    raise Exception.CreateFmt('CoInitialize failed (0x%.8x)', [OleResult]);
    { ^ doesn't use a SetupMessage since messages probably aren't loaded
      during 'initialization' section below, which calls this procedure }
end;

initialization
  InitOle;
finalization
  CoUninitialize;
{$ENDIF}
end.
